"""
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by 
the Free Software Foundation, either version 3 of the License, 
or (at your option) any later version.


This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.


You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.

Copyright © 2019 Cloud Linux Software Inc.

This software is also available under ImunifyAV commercial license,
see <https://www.imunify360.com/legal/eula>
"""
import inspect

from defence360agent.contracts.config import SecureSiteConfig
from defence360agent.feature_management.control import (
    is_native_feature_management_enabled,
)
from defence360agent.rpc_tools.lookup import wraps
from defence360agent.rpc_tools.validate import ValidationError


def secure_site_blocks_feature_management(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        if SecureSiteConfig.ENABLED:
            raise ValidationError(
                "Command is disabled because Secure Site feature is active."
            )
        return await f(*args, **kwargs)
    return wrapper


def builtin_feature_management_only(cls):
    """
    This decorator is intended to wrap rpc endpoint classes. It will
    throw ValidationError if native feature management is enabled
    when any of decorated class methods is called
    """
    def _wrapper(coro):
        @wraps(coro)
        @secure_site_blocks_feature_management
        async def wrapper(*args, **kwargs):
            if await is_native_feature_management_enabled():
                raise ValidationError("Command is disabled because native "
                                      "feature management is enabled. "
                                      "Please use your hosting panel interface"
                                      " to manage features")
            return await coro(*args, **kwargs)
        return wrapper

    assert inspect.isclass(cls), 'This decorator can only be used for classes'

    for m_name, m_obj in getattr(cls, '__dict__', {}).items():
        if not m_name.startswith('_') and \
                inspect.iscoroutinefunction(m_obj):
            wrapped = _wrapper(m_obj)
            setattr(cls, m_name, wrapped)
    return cls
