"""
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by 
the Free Software Foundation, either version 3 of the License, 
or (at your option) any later version.


This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.


You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.

Copyright © 2019 Cloud Linux Software Inc.

This software is also available under ImunifyAV commercial license,
see <https://www.imunify360.com/legal/eula>
"""
import logging
from peewee import SqliteDatabase
import threading


class OverridingReset(Exception):
    """
    Overriding reset could be a signal of logic error
    thus need to be explicitly handled in all places where
    this exception is expected to occur.
    """
    pass


logger = logging.getLogger(__name__)
_thread_local_storage = threading.local()


class SqliteDatabaseWrapper(SqliteDatabase):
    def execute_sql(self, *args, **kwargs):
        _validate(*args, **kwargs)
        return super().execute_sql(*args, **kwargs)


def reset(new_value=None):
    if hasattr(_thread_local_storage, 'thread_ident_memo'):
        raise OverridingReset()

    _thread_local_storage.thread_ident_memo = \
        new_value or threading.get_ident()


def _validate(*args, **kwargs):
    thread_ident_memo = getattr(_thread_local_storage,
                                'thread_ident_memo',
                                None)

    if thread_ident_memo is None:
        logger.error("wrong thread or _validate() was not preceded by reset()")

    elif (thread_ident_memo != threading.get_ident()):
        logger.error("thread_ident_memo check failed [%r != %r]\n"
                     "context:\nargs: %s\nkwargs: %s",
                     thread_ident_memo,
                     threading.get_ident(),
                     args,
                     kwargs)
