"""
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by 
the Free Software Foundation, either version 3 of the License, 
or (at your option) any later version.


This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.


You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.

Copyright © 2019 Cloud Linux Software Inc.

This software is also available under ImunifyAV commercial license,
see <https://www.imunify360.com/legal/eula>

Notify systemd about process state"""
import logging
import os

from defence360agent.contracts.config import ANTIVIRUS_MODE


logger = logging.getLogger(__name__)


class AgentState(object):
    """Allowed agent state for notifying systemd."""

    READY = "READY=1"
    STARTING = "STATUS=Starting main process"
    MIGRATING = "STATUS=Applying database migrations"
    DAEMONIZED = "STATUS=Demonized"


def notify(state):
    """
    Send notification to systemd, allowed formats described here
    https://www.freedesktop.org/software/systemd/man/sd_notify.html

    For example:

        notify("STATUS=Almost ready")

    """
    if ANTIVIRUS_MODE:
        return
    try:
        import sdnotify
    # We've made this import inside function deliberately because some
    # systems don't have systemd and stdnotify thus
    except ImportError as e:
        logger.info("Can't import sdnotify, reason: %s", e)
    else:
        if "NOTIFY_SOCKET" not in os.environ:
            logger.warning(
                "$NOTIFY_SOCKET environment variable is not set. It is set by"
                " service manager for supervised processes for status and"
                " start-up completion notification. Nowhere to send the"
                " status."
            )
        else:
            try:
                n = sdnotify.SystemdNotifier(debug=True)
                n.notify(state)

            except Exception as e:  # NOSONAR
                logger.exception(
                    "some problem has occurred during notifying of"
                    " systemd: %s",
                    e,
                )
