"""
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by 
the Free Software Foundation, either version 3 of the License, 
or (at your option) any later version.


This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.


You should have received a copy of the GNU General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.

Copyright © 2019 Cloud Linux Software Inc.

This software is also available under ImunifyAV commercial license,
see <https://www.imunify360.com/legal/eula>
"""
import re
from logging import getLogger

from defence360agent.contracts.config import Malware as Config
from defence360agent.contracts.license import LicenseError
from defence360agent.contracts.messages import MessageType
from defence360agent.contracts.plugins import (
    MessageSink,
    MessageSource,
    expect,
)
from defence360agent.malwarelib.utils import malware_response

logger = getLogger(__name__)


class SuspiciousToMRSUploader(MessageSink, MessageSource):
    ERR_MSG = "Failed to submit a suspicious"
    SUSP_PATTERN = re.compile(r"(?:suspicious\..+|[CS]MW-SUS-.+|SMW-HEUR-ELF)")

    async def create_source(self, loop, sink):
        self._sink = sink

    async def create_sink(self, loop):
        pass

    @classmethod
    def _check_hit(cls, hit):
        return hit["suspicious"] and cls.SUSP_PATTERN.match(hit["matches"])

    @expect(MessageType.MalwareScan)
    async def process_scan(self, message):
        results = message["results"]
        if results is None:
            return
        if not Config.SEND_FILES:
            logger.info("Uploading of suspicious files to MRS is disabled")
            return
        suspicious_hits = [
            file
            for file, data in results.items()
            if any(self._check_hit(hit) for hit in data["hits"])
        ]
        if suspicious_hits:
            # Move uploading to another task
            await self._sink.process_message(
                MessageType.MalwareMRSUpload(
                    {"suspicious_hits": suspicious_hits}
                )
            )

        errors = message["summary"].get("errors")
        if errors:
            error_hits = [hit["file"] for hit in errors]
            await self._sink.process_message(
                MessageType.MalwareMRSUpload(
                    suspicious_hits=error_hits, upload_reason="scan_error"
                )
            )

    @expect(MessageType.MalwareMRSUpload)
    async def process_hits(self, message):
        suspicious_hits = message["suspicious_hits"]
        upload_reason = message.get("upload_reason", "suspicious")
        for file in suspicious_hits:
            try:
                await malware_response.upload_with_retries(
                    file,
                    upload_reason=upload_reason,
                )
            except LicenseError as e:
                logger.warning("Cannot process message %s: %s", message, e)
                break
            except FileNotFoundError as e:
                err = "{}. {}".format(self.ERR_MSG, e.strerror)
                logger.warning("%s: %s", err, e.filename)
